.nolist
include \masm32\include\masm32rt.inc
include \masm32\macros\ucmacros.asm
include \masm32\include\shlwapi.inc
includelib \masm32\lib\shlwapi.lib

.686
.xmm

;
;   CPU architecture structure.
;
STRU_CPU_ARCH               STRUCT
                            wCPU_Arch               WORD ?
                            wReserved               WORD ?
STRU_CPU_ARCH               ENDS

;
;   System Info structure.
;
STRU_SYS_INFO               STRUCT
                                UNION               CPU_TYPE
                                    dOEMID          DWORD ?
                                    CPUArchiture    STRU_CPU_ARCH <>
                                ENDS
                            dPageSize               DWORD ?
                            pMinAppAddress          PDWORD ?
                            pMaxAppAddress          PDWORD ?
                            pActiveCPUMask          PDWORD ?
                            dNumberOfCPUS           DWORD ?
                            dCPUType                DWORD ?
                            dAllocationGramularity  DWORD ?
                            dProcessorLevel         WORD ?
                            dProcessorRevision      WORD ?
STRU_SYS_INFO               ENDS

.data


ALIGN 16
    cbDummy         BYTE 1                  ;   Force an unaligned string.
    wCopySrc        WORD 110 DUP (0)
ALIGN 16

.data?

    dVirtualBase    DWORD ?
    dSize           DWORD ?
    System_Info     STRU_SYS_INFO <>
    dPageSize       DWORD ?
    dSystemStatus   DWORD ?
    pwSrc           PWORD ?
    pwX             PWORD ?
ALIGN 16
    pwSrcCopy       DWORD (65536+16) DUP (?)
ALIGN 16
    pwPatCopy       DWORD (65536+16) DUP (?)

    WordAlign       PROTO,
                    pwDest:PWORD,
                    pwSrc:PWORD

.code

;-------------------------------------------------------------------------------
;
;   This first test will create two aligned wide character strings (one a single
;   wide character) and will then copy the first string to an unaligned
;   location, then compare the first string and the copy (aligned to unaligned).
;
;-------------------------------------------------------------------------------

start:

	mov pwSrc, uni$("my other brother darryl my other brother darryl my other brother darryl my other brother darryl xxxx")
    mov pwX,   uni$("x")
    INVOKE crt_wcscpy,OFFSET wCopySrc,pwSrc
    INVOKE crt_wcscmp,OFFSET wCopySrc,pwSrc
    or     eax,eax
    jz     Ok1
mov edi,1
int 3
Ok1:
;int 3
;
;   I checked here with an int 3 - the copy was correct, on an unaligned bound.
;

;-------------------------------------------------------------------------------
;
;   This second test will search the unaligned string for a single character.
;
;-------------------------------------------------------------------------------

    mov    eax,pwX
    movzx  ebx,WORD PTR [eax]
    INVOKE crt_wcschr,OFFSET wCopySrc,ebx
    or     eax,eax
    jnz    Ok2
mov edi,2
int 3
Ok2:
;int 3
;
;   I checked here with an int 3 - the first 'x' was found in the unaligned
;   string.
;

;-------------------------------------------------------------------------------
;
;   wcs_ functions are not alignment sensitive! But, non-alignment can cause
;   problems for SSE implementations because aligned loads of SSE regs will get
;   wide characters split across word boundaries in the reg such that pcmpeqw
;   will not work correctly!
;
;   Catch 22! You need aligned SSE reg loads to prevent buffer overrun, but
;   aligned loads will cause character mis-alignment in the SSE reg!
;
;   All hope is not lost, see below.
;
;-------------------------------------------------------------------------------

;
;   Get system info
;
    INVOKE GetSystemInfo,OFFSET System_Info
    mov    eax,System_Info.dPageSize
    mov    dPageSize,eax
;
;   Calculate the space required modulo page size.
;
;
;   Reserve the space required.
;
    INVOKE VirtualAlloc,0,dPageSize,MEM_RESERVE,PAGE_NOACCESS
    mov    dVirtualBase,eax
    or     eax,eax
    jnz    Ok3
mov edi,3
int 3
Ok3:
;
;   Commit the space required.
;
    INVOKE VirtualAlloc,dVirtualBase,dPageSize,MEM_COMMIT,PAGE_READWRITE
    or     eax,eax
    jnz    Ok4
mov edi,4
int 3
Ok4:

;-------------------------------------------------------------------------------
;
;   Test for aligned short string at the end of the buffer - this works.
;
;-------------------------------------------------------------------------------

GoodTestAligned:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.
    add    ebx,dPageSize                    ;   Get the buffer end.
    sub    ebx,7                            ;   Get the string start (6 NULLS and a NULL).
;
;   This would be the start of the function with string start in ebx.
;
    mov         ecx,ebx                     ;   Save the string start.
    and         ebx,-16                     ;   Force mod 16 alignment.
    mov         edx,0FFFFh                  ;   Get full mask.
    pxor   xmm1,xmm1                        ;   Get 16 nulls.
    sub         ecx,ebx                     ;   Calculate leading garbage characters.
    movdqa xmm0,OWORD PTR [ebx]             ;   Get start of string.
    shl         edx,cl                      ;   Correct the mask for the leading garbage.
    pcmpeqb xmm0,xmm1                       ;   Calculate matches.
    lea         ebx,[ebx+16]                ;   Set for next aligned load.
    pmovmskb ecx,xmm0                       ;   Extract bitmap of matches.
    and         ecx,edx                     ;   Correct the bitmap for garbage the first time.
    mov         edx,0FFFFh                  ;   Force all remaining checks to be full checks.
    bsf         ecx,ecx                     ;   Get the first match BYTE position.
;
;   ...
;

;-------------------------------------------------------------------------------
;
;   Enter this to test scanning a wide character string starting at an odd BYTE.
;
;   Note: It is always assumed that the wide character strings are properly null
;   terminated and lie within the buffer.
;
;-------------------------------------------------------------------------------

    mov    ecx,201                          ;   Set the string size plus null plus 1.

GoodTestUnalignedWide:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.
    add    ebx,dPageSize                    ;   Get the buffer end.
    sub    ebx,ecx                          ;   Get the string start.
    mov    esi,ebx                          ;   Use string as the buffer start if not aligning.
    mov    edi,pwSrc
    mov    WORD PTR [edi+ecx-3],0           ;   Truncate the source string at string size.

    pushad
    INVOKE crt_wcscpy,ebx,edi
    popad

    test   ebx,1
    jz     DoPat
    mov    esi,OFFSET pwSrcCopy             ;   Save the aligned start.
    INVOKE WordAlign,esi,ebx                ;   Word align the wide characters in a guaranteed long buffer.
    test   eax,0Fh                          ;   Valid return?
    jz     Ok5                              ;   Yes
mov edi,5
int 3
Ok5:

    pushad
    INVOKE crt_wcscmp,esi,ebx
    cmp    eax,0
    popad
    jz     Ok6

mov edi,6
int 3
Ok6:
;
;   String done, get the Pattern.
;
DoPat:
    mov    edx,dVirtualBase                 ;   Get the buffer start.
    add    edx,dPageSize                    ;   Get the buffer end.
    sub    edx,ecx                          ;   Get the string start.
    mov    edi,edx                          ;   Use string as the buffer start if not aligning.

    pushad
    INVOKE crt_wcscpy,edx,pwSrc
    popad

    test   edx,1
    jz     IsAligned
    mov    edi,OFFSET pwPatCopy             ;   Save the aligned start.
    INVOKE WordAlign,edi,edx                ;   Word align the wide characters in a guaranteed long buffer.
    test   eax,0Fh                          ;   Valid return?
    jz     Ok7                              ;   Yes
mov edi,7
int 3
Ok7:

    pushad
    INVOKE crt_wcscmp,edi,edx
    cmp    eax,0
    popad
    jz     Ok8

mov edi,8
int 3
Ok8:
;
;   esi edi are aligned for compares, or whatever. 
;
IsAligned:
;
;   ...
;
    lea    ecx,[ecx-2]                      ;   Test all string lengths down to 3.
    cmp    ecx,3
    ja     GoodTestUnalignedWide

;-------------------------------------------------------------------------------
;
;   Test for short string at the end of the buffer - this does not work.
;
;   This will fail for either aligned or mis-aligned strings.
;
;   The failure will an "access violation reading..."  for a load of xmm7.
;
;-------------------------------------------------------------------------------

BadTest:
    mov    ebx,dVirtualBase                 ;   Get the buffer start.

    add    ebx,dPageSize                    ;   Get the buffer end.
    sub    ebx,7                            ;   Get the string start (6 NULLS and a NULL).
;
;   This would be the start of the function with string start in ebx.
;
    movdqu xmm7,OWORD PTR [ebx]             ;   This will fail right here.

	exit

;-------------------------------------------------------------------------------
;
;   Align an unaligned wide character string in a buffer, checking for nulls.
;   Unaligned in this case means the wide chararacter string starts on an odd
;   BYTE bound. The buffer is expected to be mod 16 bounded, and 65536+16 bytes
;   long (maximum supported string length + 16 byte pad to prevent overrun).
;
;   Returns the pointer to the buffer, or returns a flag if an error is detected
;   (1 if wide character string too long, 2 if buffer is not mod 16 bounded, 3
;   if not an unaligned wide character string), preserves all regs except eax.
;
;   Test the return for least significant 4 bits to check validity (should all
;   be 0 if valid)
;
;-------------------------------------------------------------------------------

WordAlign   PROC USES ebx ecx esi edi,
            pwBuf:PWORD,                    ;   Aligned buffer for copy. 
            pwStr:PWORD                     ;   Unaligned wide string.

    mov    ebx,pwStr                        ;   Get the unaligned string pointer.
    mov    edi,pwBuf                        ;   Get the aligned buffer pointer.
    mov    eax,1                            ;   Set pwStr not odd error return.
    test   ebx,1                            ;   pwStr valid?
    jz     Exit                             ;   No.
    mov    eax,2                            ;   Set pwBuf not mod 16 error return.
    test   edi,0Fh                          ;   pwBuf valid?
    jnz    Exit                             ;   No.
    mov    esi,ebx                          ;   Save pwStr for DoUnaligned.
    mov    ecx,ebx                          ;   Save the string start.
    and    ebx,-16                          ;   Force mod 16 alignment.
    lea    eax,[edi+65536]                  ;   Set first invalid address for DoUnaligned.
    sub    ecx,ebx                          ;   Calculate leading garbage BYTES.
    cmp    ecx,15                           ;   No valid WORDS (only half of one WORD for the xmm reg)?
    jz     DoUnaligned                      ;   Yes, just use unaligned loads.
    lea    ecx,[ecx-1]                      ;   Decrement ecx for the alignment shift.
    pxor        xmm1,xmm1                   ;   Get 16 nulls.
    movdqa      xmm0,OWORD PTR [ebx]        ;   Get aligned data containing the start of the string.
    psrldq      xmm0,1                      ;   WORD align the wide characters left (discard first BYTE).
    movdqa      xmm2,xmm0                   ;   Copy the aligned data.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches on remaining 7 WORDS.
    pmovmskb    eax,xmm0                    ;   Extract bitmap (16 bits for 8 WORDS) of null matches.
    shr    ax,cl                            ;   Mask the matches for valids.
    test   ecx,2                            ;   Test the WORD shift count for 1 (ecx is BYTES here).
    jz     SkipOne                          ;   Not shift 1.
    psrldq      xmm2,2                      ;   Shift the wide characters left 1 WORD.
SkipOne:
    test   ecx,4                            ;   Test the WORD shift count for 2.
    jz     SkipTwo                          ;   Not shift 2.
    psrldq      xmm2,4                      ;   Shift the wide characters left 2 WORDS.
SkipTwo:
    test   ecx,8                            ;   Test the WORD shift count for 4.
    jz     SkipThree                        ;   Not shift 4.
    psrldq      xmm2,8                      ;   Shift the wide characters left 4 WORDS.
SkipThree:
    movdqa      OWORD PTR [edi],xmm2        ;   Save the string head.
    mov    ebx,14                           ;   Total bytes that could have been in xmm0.
    sub    ebx,ecx                          ;   Total valid bytes that remain.
    lea    esi,[esi+ebx]                    ;   Set address for next unaligned load.
    lea    edi,[edi+ebx]                    ;   Set address for next unaligned store.
    test   eax,eax                          ;   Any valid null matches found?
    jnz    Good                             ;   Yes, go no further.
    mov    eax,pwBuf                        ;   Point to the buffer.
    lea    eax,[eax+65536]                  ;   Set the first invalid address.
;
;   Move unaligned data to aligned buffer.
;
DoUnaligned:
    movdqu      xmm0,OWORD PTR [esi]        ;   Get unaligned data, guaranteed no buffer overrun.
    movdqu      OWORD PTR [edi],xmm0        ;   Build an unaligned buffer, guaranteed no buffer overrun.
    pcmpeqw     xmm0,xmm1                   ;   Calculate null matches.
    pmovmskb    ecx,xmm0                    ;   Extract bitmap of null matches.
    lea    edi,[edi+16]                     ;   Set up for next store.
    lea    esi,[esi+16]                     ;   Set up for next load.
    cmp    edi,eax                          ;   Valid store address?
    ja     Overflow                         ;   No.
    test   ecx,ecx                          ;   Any nulls?
    jz     DoUnaligned                      ;   No, keep copying.
;
;   Set good return (pointer to buffer - pwBuf).
;
Good:
    mov    eax,pwBuf                        ;   Set good return.
    jmp    Exit
;
;   Set buffer overflow error return.
;
Overflow:
    mov    eax,3
    jmp    Exit
;
;   Exit WordAlign.
;
Exit:
    ret    8
WordAlign   ENDP

end start

